# python3

import unittest
import os
import socket
import getpass

from .embedded_server.openssh import OpenSSHServer
from ssh2.session import Session


BASE_PATH = os.path.abspath(os.path.dirname(__file__))


class SSH2TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.user = getpass.getuser()
        cls.user_key = os.path.join(BASE_PATH, 'unit_test_key')
        cls.user_ec_key = os.path.join(BASE_PATH, 'key_ecdsa')
        os.chmod(cls.user_key, 0o600)
        os.chmod(cls.user_ec_key, 0o600)
        cls.user_pub_key = f"{cls.user_key}.pub"
        cls.user_ec_pub_key = f"{cls.user_ec_key}.pub"
        cls.server = OpenSSHServer()
        cls.server.start_server()

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()
        del cls.server

    def setUp(self):
        self.host = '127.0.0.1'
        self.port = 2222
        self.cmd = 'echo me'
        self.resp = 'me'
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.host, self.port))
        self.sock = sock
        self.session = Session()
        self.session.handshake(self.sock)

    def tearDown(self):
        del self.session
        self.sock.close()
        del self.sock

    def _auth(self):
        return self.session.userauth_publickey_fromfile(self.user, self.user_key)


class SSH2CertTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.user = getpass.getuser()
        cls.user_key = os.path.join(BASE_PATH, 'signed_key_ecdsa')
        cls.user_pub_key = f"{cls.user_key}.pub"
        cls.user_cert = f"{cls.user_key}-cert.pub"
        cls.server = OpenSSHServer()
        cls.server.sign_key(cls.user)
        cls.server.start_server()

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()
        del cls.server

    def setUp(self):
        self.host = '127.0.0.1'
        self.port = 2222
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.host, self.port))
        self.sock = sock
        self.session = Session()
        self.session.handshake(self.sock)

    def tearDown(self):
        del self.session
        self.sock.close()
        del self.sock



# This file is part of ssh2-python.
# Copyright (C) 2017 Panos Kittenis

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import os
import socket
import subprocess
from time import sleep

DIR_NAME = os.path.dirname(__file__)
PDIR_NAME = os.path.dirname(DIR_NAME)
PPDIR_NAME = os.path.dirname(PDIR_NAME)


class OpenSSHServer(object):

    def __init__(self, port=2222):
        self.port = port
        self.server_proc = None
        self.server_key = os.path.join(DIR_NAME, "rsa_host_key")
        self.sshd_config = os.path.join(DIR_NAME, "sshd_config")
        self.authorized_keys = os.path.join(DIR_NAME, "authorized_keys")
        self.ca_key = os.path.join(DIR_NAME, "ca.pub")
        self._fix_permissions()
        self.cmdline = ['/usr/sbin/sshd', '-D', '-q', '-p', str(port),
                        '-o', f'AuthorizedKeysFile={self.authorized_keys}',
                        '-o', f'TrustedUserCAKeys={self.ca_key}',
                        '-h', str(self.server_key), '-f', str(self.sshd_config)]

    def _fix_permissions(self):
        os.chmod(self.server_key, 0o600)
        for _dir in [DIR_NAME, PDIR_NAME, PPDIR_NAME]:
            os.chmod(_dir, 0o755)

    def start_server(self):
        # print("Running:", " ".join(self.cmdline), file=sys.stderr)
        server = subprocess.Popen(self.cmdline, shell=False, stdout=subprocess.DEVNULL)
        self.server_proc = server
        self._wait_for_port()

    def _wait_for_port(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sleep(.1)
        while sock.connect_ex(('127.0.0.1', self.port)) != 0:
            sleep(.1)
        sock.close()

    @classmethod
    def sign_key(cls, user):
        os.chmod(f'{DIR_NAME}/ca', 0o600)
        cmd = ['ssh-keygen', '-s', f'{DIR_NAME}/ca', '-I', 'myidentity',
               '-n', user, '-V', '+1w', '-z', '1', f'{PDIR_NAME}/signed_key_ecdsa.pub']
        subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    def stop(self):
        if self.server_proc is not None and self.server_proc.returncode is None:
            self.server_proc.terminate()
            return self.server_proc.wait()

    def __del__(self):
        self.stop()



r"""Test alternate authentication methods.

As done in the libssh2.

Only the following key types are supported by libssh2 at this time:

    - ecdsa-sha2-nistp256-cert-v01@openssh.com
    - ecdsa-sha2-nistp384-cert-v01@openssh.com
    - ecdsa-sha2-nistp521-cert-v01@openssh.com

Preferred key type:
    ecdsa-sha2-nistp384-cert-v01@openssh.com

Generate ca_ecdsa:
ssh-keygen -C CA -f embedded_server/ca -b 521 -t ecdsa

Generated the test keypair key_ecdsa, used with non-cert test, with:
ssh-keygen -t ecdsa -b 384 -f ./key_ecdsa -C ""

Generate keypair to be signed, use with cert test:
ssh-keygen -t ecdsa -b 384 -f ./signed_key_ecdsa -C ""

Sign the keypair with the CA key generated earlier:
ssh-keygen -s ./embedded_server/ca -I myidentity -n $USER -V "always:forever" -z 1 \
        ./signed_key_ecdsa.pub

NOTE: the above step must have the principle match the user logging in. The test suite will create
      the cert file.
"""

import os

from .base_test import SSH2TestCase, SSH2CertTestCase

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


class BasicPubkeyAuthTestCase(SSH2TestCase):

    def test_ec_auth(self):
        self.assertEqual(self.session.userauth_publickey_fromfile(
                         self.user, self.user_ec_key, passphrase="",
                         publickey=self.user_ec_pub_key), 0)


class FileCertAuthTestCase(SSH2CertTestCase):

    def _auth(self):
        return self.session.userauth_publickey_fromfile(self.user, self.user_key, passphrase="",
                                                        publickey=self.user_cert)

    def test_auth(self):
        self.assertEqual(self._auth(), 0)


class MemoryCertAuthTestCase(SSH2CertTestCase):

    def _auth(self):
        with open(self.user_key, "rb") as fo:
            privatekeyfiledata = fo.read()

        with open(self.user_cert, "rb") as fo:
            publickeyfiledata = fo.read()

        return self.session.userauth_publickey_frommemory(
            self.user, privatekeyfiledata, passphrase='', publickeyfiledata=publickeyfiledata)

    def test_auth(self):
        self.assertEqual(self._auth(), 0)



# python3.6

import os
from glob import glob

import versioneer
from setuptools import setup, find_packages

from Cython.Distutils.extension import Extension
from Cython.Distutils import build_ext


LIBC = os.confstr('CS_GNU_LIBC_VERSION').replace(" ", "")

sources = glob('ssh2/*.pyx')
_libs = ['ssh2', 'ssl', 'crypto', 'z']

_fwd_default = 0
_comp_args = ["-O3"]
_have_agent_fwd = bool(int(os.environ.get('HAVE_AGENT_FWD', _fwd_default)))

cython_directives = {
    'embedsignature': True,
    'boundscheck': False,
    'optimize.use_switch': True,
    'wraparound': False,
    'language_level': "3",
}

cython_args = {
    'include_path': ["ssh2"],
    'cython_directives': cython_directives,
    'cython_compile_time_env': {
        'EMBEDDED_LIB': True,
        'HAVE_AGENT_FWD': _have_agent_fwd,
    }
}


_lib_dir = os.path.abspath(f"./build/{LIBC}/src")
include_dirs = ["libssh2/include"]

extensions = [
    Extension(pyx.split('.')[0].replace(os.path.sep, '.'),
              sources=[pyx],
              include_dirs=include_dirs,
              libraries=_libs,
              library_dirs=[_lib_dir],
              extra_compile_args=_comp_args,
              **cython_args)
    for pyx in sources]

package_data = {'ssh2': ['*.pxd']}

cmdclass = versioneer.get_cmdclass()
cmdclass['build_ext'] = build_ext

setup(
    name='ssh2-python3',
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    packages=find_packages(
        '.', exclude=('embedded_server', 'embedded_server.*',
                      'tests', 'tests.*',
                      '*.tests', '*.tests.*')),
    zip_safe=False,
    include_package_data=True,
    tests_require=['pytest'],
    python_requires='~=3.6',
    license='LGPLv2',
    author='Panos Kittenis',
    author_email='22e889d8@opayq.com',
    maintainer='Keith Dart',
    maintainer_email='keith.dart@gmail.com',
    description='Super fast SSH library - bindings for libssh2 and Python 3',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/pycopia/ssh2-python3',
    platforms=['linux_x86_64', 'manylinux2014_x86_64 '],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3 :: Only",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System :: Shells',
        'Topic :: System :: Networking',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
    ],
    ext_modules=extensions,
    package_data=package_data,
)
